#Errores
class ErrorBDD(Exception):...

class ErrorMalaSolicitud(ErrorBDD):
        """Excepción para errores relacionados con solicitudes"""
class SinResultado(ErrorMalaSolicitud): ...
class ErrorTablaNoExiste(ErrorBDD): ...
class ErrorBaseDeDatosNoExiste(ErrorBDD): ...

class ErrorPoolLlena(ErrorBDD): ...
class ErrorDemasiadasConexiones(ErrorBDD): ...

class IdsNoCoinciden(ErrorBDD): ...

class ErrorAPI(Exception): ...